@@ -1,5 +1,8 @@
 #include "mupdf/fitz.h"
 
+#include <string.h>
+#include <limits.h>
+
 /* Fax G3/G4 decoder */
 
 /* TODO: uncompressed */
@@ -18,13 +21,11 @@
 <raph> peter came up with this, and it makes sense
 */
 
-typedef struct cfd_node_s cfd_node;
-
-struct cfd_node_s
+typedef struct
 {
 	short val;
 	short nbits;
-};
+} cfd_node;
 
 enum
 {
@@ -291,8 +292,6 @@ static inline void setbits(unsigned char *line, int x0, int x1)
 	}
 }
 
-typedef struct fz_faxd_s fz_faxd;
-
 enum
 {
 	STATE_INIT,		/* initial state, optionally waiting for EOL */
@@ -303,7 +302,7 @@ enum
 	STATE_DONE		/* all done */
 };
 
-struct fz_faxd_s
+typedef struct
 {
 	fz_stream *chain;
 
@@ -329,7 +328,7 @@ struct fz_faxd_s
 	unsigned char *rp, *wp;
 
 	unsigned char buffer[4096];
-};
+} fz_faxd;
 
 static inline void eat_bits(fz_faxd *fax, int nbits)
 {
@@ -766,24 +765,19 @@ close_faxd(fz_context *ctx, void *state_)
 	fz_free(ctx, fax);
 }
 
-/* Default: columns = 1728, end_of_block = 1, the rest = 0 */
 fz_stream *
 fz_open_faxd(fz_context *ctx, fz_stream *chain,
 	int k, int end_of_line, int encoded_byte_align,
 	int columns, int rows, int end_of_block, int black_is_1)
 {
-	fz_faxd *fax = NULL;
+	fz_faxd *fax;
 
-	fz_var(fax);
+	if (columns < 0 || columns >= INT_MAX - 7)
+		fz_throw(ctx, FZ_ERROR_GENERIC, "too many columns lead to an integer overflow (%d)", columns);
 
+	fax = fz_malloc_struct(ctx, fz_faxd);
 	fz_try(ctx)
 	{
-		if (columns < 0 || columns >= INT_MAX - 7)
-			fz_throw(ctx, FZ_ERROR_GENERIC, "too many columns lead to an integer overflow (%d)", columns);
-
-		fax = fz_malloc_struct(ctx, fz_faxd);
-		fax->chain = chain;
-
 		fax->ref = NULL;
 		fax->dst = NULL;
 
@@ -806,23 +800,21 @@ fz_open_faxd(fz_context *ctx, fz_stream *chain,
 		fax->dim = fax->k < 0 ? 2 : 1;
 		fax->eolc = 0;
 
-		fax->ref = fz_malloc(ctx, fax->stride);
-		fax->dst = fz_malloc(ctx, fax->stride);
+		fax->ref = Memento_label(fz_malloc(ctx, fax->stride), "fax_ref");
+		fax->dst = Memento_label(fz_malloc(ctx, fax->stride), "fax_dst");
 		fax->rp = fax->dst;
 		fax->wp = fax->dst + fax->stride;
 
 		memset(fax->ref, 0, fax->stride);
 		memset(fax->dst, 0, fax->stride);
+
+		fax->chain = fz_keep_stream(ctx, chain);
 	}
 	fz_catch(ctx)
 	{
-		if (fax)
-		{
-			fz_free(ctx, fax->dst);
-			fz_free(ctx, fax->ref);
-		}
+		fz_free(ctx, fax->dst);
+		fz_free(ctx, fax->ref);
 		fz_free(ctx, fax);
-		fz_drop_stream(ctx, chain);
 		fz_rethrow(ctx);
 	}
 
