@@ -1,9 +1,6 @@
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
-#include "../fitz/font-imp.h"
-#include "../fitz/fitz-imp.h"
-
 #include <assert.h>
 
 #include <ft2build.h>
@@ -139,7 +136,7 @@ static int strcmp_ignore_space(const char *a, const char *b)
 const char *pdf_clean_font_name(const char *fontname)
 {
 	int i, k;
-	for (i = 0; i < nelem(base_font_names); i++)
+	for (i = 0; i < (int)nelem(base_font_names); i++)
 		for (k = 0; base_font_names[i][k]; k++)
 			if (!strcmp_ignore_space(base_font_names[i][k], fontname))
 				return base_font_names[i][0];
@@ -249,17 +246,15 @@ static int ft_width(fz_context *ctx, pdf_font_desc *fontdesc, int cid)
 {
 	int mask = FT_LOAD_NO_SCALE | FT_LOAD_NO_HINTING | FT_LOAD_NO_BITMAP | FT_LOAD_IGNORE_TRANSFORM;
 	int gid = ft_cid_to_gid(fontdesc, cid);
-	FT_Fixed adv;
+	FT_Fixed adv = 0;
 	int fterr;
 	FT_Face face = fontdesc->font->ft_face;
 	FT_UShort units_per_EM;
 
 	fterr = FT_Get_Advance(face, gid, mask, &adv);
-	if (fterr)
-	{
-		fz_warn(ctx, "freetype advance glyph (gid %d): %s", gid, ft_error_string(fterr));
-		return 0;
-	}
+	if (fterr && fterr != FT_Err_Invalid_Argument)
+		fz_warn(ctx, "FT_Get_Advance(%d): %s", gid, ft_error_string(fterr));
+
 	units_per_EM = face->units_per_EM;
 	if (units_per_EM == 0)
 		units_per_EM = 2048;
@@ -290,7 +285,7 @@ static const struct { int code; const char *name; } mre_diff_table[] =
 static int lookup_mre_code(const char *name)
 {
 	int i;
-	for (i = 0; i < nelem(mre_diff_table); ++i)
+	for (i = 0; i < (int)nelem(mre_diff_table); ++i)
 		if (!strcmp(name, mre_diff_table[i].name))
 			return mre_diff_table[i].code;
 	for (i = 0; i < 256; i++)
@@ -299,6 +294,33 @@ static int lookup_mre_code(const char *name)
 	return -1;
 }
 
+static int ft_find_glyph_by_unicode_name(FT_Face face, const char *name)
+{
+	int unicode, glyph;
+
+	/* Prefer exact unicode match if available. */
+	unicode = fz_unicode_from_glyph_name_strict(name);
+	if (unicode > 0)
+	{
+		glyph = ft_char_index(face, unicode);
+		if (glyph > 0)
+			return glyph;
+	}
+
+	/* Fall back to font glyph name if we can. */
+	glyph = ft_name_index(face, name);
+	if (glyph > 0)
+		return glyph;
+
+	/* Fuzzy unicode match as last attempt. */
+	unicode = fz_unicode_from_glyph_name(name);
+	if (unicode > 0)
+		return ft_char_index(face, unicode);
+
+	/* Failed. */
+	return 0;
+}
+
 /*
  * Load font files.
  */
@@ -679,7 +701,8 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		else
 			fz_warn(ctx, "freetype could not find any cmaps");
 
-		etable = fz_malloc_array(ctx, 256, sizeof(unsigned short));
+		/* FIXME: etable may leak on error. */
+		etable = Memento_label(fz_malloc_array(ctx, 256, unsigned short), "cid_to_gid");
 		fontdesc->size += 256 * sizeof(unsigned short);
 		for (i = 0; i < 256; i++)
 		{
@@ -713,7 +736,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 						item = pdf_array_get(ctx, diff, i);
 						if (pdf_is_int(ctx, item))
 							k = pdf_to_int(ctx, item);
-						if (pdf_is_name(ctx, item) && k >= 0 && k < nelem(estrings))
+						if (pdf_is_name(ctx, item) && k >= 0 && k < (int)nelem(estrings))
 							estrings[k++] = pdf_to_name(ctx, item);
 					}
 				}
@@ -766,11 +789,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 				{
 					if (estrings[i])
 					{
-						int unicode = fz_unicode_from_glyph_name(estrings[i]);
-						if (unicode > 0)
-							glyph = ft_char_index(face, unicode);
-						else
-							glyph = ft_name_index(face, estrings[i]);
+						glyph = ft_find_glyph_by_unicode_name(face, estrings[i]);
 						if (glyph > 0)
 							etable[i] = glyph;
 					}
@@ -785,9 +804,10 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 					if (estrings[i])
 					{
 						int mrcode = lookup_mre_code(estrings[i]);
+						glyph = 0;
 						if (mrcode > 0)
 							glyph = ft_char_index(face, mrcode);
-						else
+						if (glyph == 0)
 							glyph = ft_name_index(face, estrings[i]);
 						if (glyph > 0)
 							etable[i] = glyph;
@@ -852,6 +872,7 @@ pdf_load_simple_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict)
 		}
 		fz_catch(ctx)
 		{
+			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 			fz_warn(ctx, "cannot load ToUnicode CMap");
 		}
 
@@ -924,7 +945,7 @@ hail_mary_cmp_key(fz_context *ctx, void *k0, void *k1)
 }
 
 static void
-hail_mary_format_key(fz_context *ctx, char *s, int n, void *key_)
+hail_mary_format_key(fz_context *ctx, char *s, size_t n, void *key_)
 {
 	fz_strlcpy(s, "(hail mary font)", n);
 }
@@ -1047,7 +1068,7 @@ load_cid_font(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_obj *encodi
 
 			len = fz_buffer_storage(ctx, buf, &data);
 			fontdesc->cid_to_gid_len = len / 2;
-			fontdesc->cid_to_gid = fz_malloc_array(ctx, fontdesc->cid_to_gid_len, sizeof(unsigned short));
+			fontdesc->cid_to_gid = Memento_label(fz_malloc_array(ctx, fontdesc->cid_to_gid_len, unsigned short), "cid_to_gid_map");
 			fontdesc->size += fontdesc->cid_to_gid_len * sizeof(unsigned short);
 			for (z = 0; z < fontdesc->cid_to_gid_len; z++)
 				fontdesc->cid_to_gid[z] = (data[z * 2] << 8) + data[z * 2 + 1];
@@ -1257,6 +1278,7 @@ pdf_load_font_descriptor(fz_context *ctx, pdf_document *doc, pdf_font_desc *font
 		}
 		fz_catch(ctx)
 		{
+			fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
 			fz_warn(ctx, "ignored error when loading embedded font; attempting to load system font");
 			if (!iscidfont && fontname != pdf_clean_font_name(fontname))
 				pdf_load_builtin_font(ctx, fontdesc, fontname, 1);
@@ -1307,9 +1329,8 @@ pdf_make_width_table(fz_context *ctx, pdf_font_desc *fontdesc)
 	}
 
 	font->width_count = n + 1;
-	font->width_table = fz_malloc_array(ctx, font->width_count, sizeof(int));
-	memset(font->width_table, 0, font->width_count * sizeof(int));
-	fontdesc->size += font->width_count * sizeof(int);
+	font->width_table = Memento_label(fz_malloc_array(ctx, font->width_count, short), "font_widths");
+	fontdesc->size += font->width_count * sizeof(short);
 
 	font->width_default = fontdesc->dhmtx.w;
 	for (i = 0; i < font->width_count; i++)
