@@ -3,6 +3,8 @@
 # Pre-generates the expected output subset files (via fonttools) for
 # specified subset test suite(s).
 
+from __future__ import print_function, division, absolute_import
+
 import io
 import os
 import sys
@@ -12,19 +14,19 @@ from subset_test_suite import SubsetTestSuite
 
 
 def usage():
-	print "Usage: generate-expected-outputs.py <test suite file> ..."
+	print("Usage: generate-expected-outputs.py <test suite file> ...")
 
 
 def generate_expected_output(input_file, unicodes, profile_flags, output_path):
 	args = ["fonttools", "subset", input_file]
-	args.extend(profile_flags)
 	args.extend(["--notdef-outline",
-		     "--name-IDs=*",
 		     "--name-languages=*",
 		     "--name-legacy",
-		     "--drop-tables+=DSIG,GPOS,GSUB,GDEF",
+                     "--layout-features=*",
+		     "--drop-tables+=DSIG,GPOS,GSUB,GDEF,gvar,avar,MVAR,HVAR",
 		     "--unicodes=%s" % unicodes,
 		     "--output-file=%s" % output_path])
+	args.extend(profile_flags)
 	check_call(args)
 
 
@@ -37,11 +39,11 @@ for path in args:
 		test_suite = SubsetTestSuite(path, f.read())
 		output_directory = test_suite.get_output_directory()
 
-		print "Generating output files for %s" % output_directory
+		print("Generating output files for %s" % output_directory)
 		for test in test_suite.tests():
 			unicodes = test.unicodes()
 			font_name = test.get_font_name()
-			print "Creating subset %s/%s" % (output_directory, font_name)
+			print("Creating subset %s/%s" % (output_directory, font_name))
 			generate_expected_output(test.font_path, unicodes, test.get_profile_flags(),
 						 os.path.join(output_directory,
 							      font_name))
